structure HornQ =

struct

   structure H = Horn

   structure I = IntSets
   structure T = Table (structure Item = IntPairItem
			type value = I.IntSet list)

(* constraints with enumerated queries *)

   datatype 'a pre = 	U of int * 'a H.ind list * I.IntSet
		|	N of int * 'a H.ind list
		|	Eq of 'a H.ind * 'a H.ind
		|	Neq of 'a H.ind * 'a H.ind
		|	FunApp of 'a H.ind * ('a list -> 'a) * 'a H.ind list
		|	And of 'a pre * 'a pre
		|	Or of 'a pre * 'a pre
		|	Exists of int * 'a pre
		|	CheapOr of 'a pre * 'a pre
		|	CheapExists of int * 'a pre
		|	Yes | No
		
   datatype 'a clause = 	One
		|	R of int * 'a H.ind list
		|	Implies of 'a pre * 'a clause
		|	Break of 'a clause	
                |	Both of 'a clause * 'a clause
		|	Forall of int * 'a clause

   fun min ([],[]) = []
    |  min ((x1,b1)::l1, (x2,b2)::l2) = if x1 = x2 then case (b1,b2) 
		of (SOME (),_) => (x1,b1):: min (l1,l2)
		|  _           => (x2,b2):: min (l1,l2)
		else raise (Fail "\n\nIncompatible var sets :-(\n\n")
    |  min _ = raise (Fail "\n\nIllegal use of ``min''\n\n")

   fun inst (H.Const _, list) = list
    |  inst (H.Var x, []) = raise (Fail "\n\nGlobal variable???\n\n")
    |  inst (H.Var y, (x,b)::rest) = if x = y then (x,SOME ()):: rest
		else (x,b) :: inst (H.Var y, rest)
		
   fun prefix list = let
       fun doit (i,[]) = true
	|  doit (i,j::rest) = if i = j then doit (i+1,rest)
			      else false
       in doit (0,list)
       end

   fun getInstSet (args, list) = let
	fun isInst (H.Const _,_) = true
	 |  isInst (H.Var x, []) = false
	 |  isInst (H.Var x, (y,b)::rest) = if x = y then case b
		of SOME () => true
		|  NONE    => false
		else isInst (H.Var x, rest)

	fun doit (i,[]) =  []
	 |  doit (i,a::aa) = if isInst (a,list) 
		then i :: doit (i+1,aa)
		else doit (i+1,aa)

	val result = doit (0,args)

	in if prefix result then I.emptyIntSet 1
	   else I.IntList2Set result
	end

(* ... annotates each positive query with the set of instantiated
   components ... *)

   fun elab clause = let

	val acTable  = T.makeTable 101
	fun addAccess ((n,k),set) = case T.sub (acTable, (n,k))
		of SOME list => T.update (acTable, (n,k), set::list)
		|  NONE      => T.update (acTable, (n,k), [set])   
		
	fun doPre H.Yes list = (Yes, list)
	 |  doPre H.No list  = (No,  list)
	 |  doPre (H.CheapExists (int, pre)) list = let
		val (pre, _::list) = doPre pre ((int,NONE)::list)
		in (CheapExists (int, pre), list)
		end
	 |  doPre (H.Exists (int, pre)) list = let
		val (pre, _::list) = doPre pre ((int,NONE)::list)
		in (Exists (int, pre), list)
		end
	 | doPre (H.CheapOr (p1,p2)) list = let
		val (p1,l1) = doPre p1 list
		val (p2,l2) = doPre p2 list
		in (CheapOr (p1,p2), min (l1,l2))
		end
	 | doPre (H.Or (p1,p2)) list = let
		val (p1,l1) = doPre p1 list
		val (p2,l2) = doPre p2 list
		in (Or (p1,p2), min (l1,l2))
		end
	 | doPre (H.And (p1,p2)) list = let
		val (p1,list) = doPre p1 list
		val (p2,list) = doPre p2 list
		in (And (p1,p2), list)
		end
	 | doPre (H.Neq (x,y)) list = 
		(Neq (x,y), list)
	 | doPre (H.Eq (x,y)) list = 
		(Eq (x,y), foldl inst list [x,y])
	 | doPre (H.FunApp (x,f,yy)) list =
		(FunApp (x,f,yy), inst (x,list))
	 | doPre (H.N (n, args)) list =
		(N (n, args), list)
	 | doPre (H.U (n, args)) list = let
		val k = List.length args
		val set = getInstSet (args, list)
		val _ = if not (I.isEmptyIntSet set) then 
			addAccess ((n,k),set)
			else ()
		in (U (n, args, set), foldl inst list args)
		end

	fun doit H.One list = One
	 |  doit (H.Forall (int,clause)) list = Forall (int, doit clause ((int,NONE)::list))
	 |  doit (H.Both (c1,c2)) list = Both (doit c1 list, doit c2 list)
	 |  doit (H.Implies (pre,c)) list = let
		val (pre, list) = doPre pre list
		in Implies (pre, doit c list)
		end
         |  doit (H.Break c) list = Break (doit c list)
	 |  doit (H.R (n, args)) list = R (n, args)

	val clause = doit clause []
	in (clause, acTable)
	end
end

