signature IntSets =
   sig 
      type IntSet

      val fullIntSet    : int -> IntSet
      val emptyIntSet   : int -> IntSet
      val singleIntSet  : int -> IntSet

      val isEmptyIntSet : IntSet -> bool
      val eqIntSets     : IntSet * IntSet -> bool 
      val super		: IntSet * IntSet -> bool
      val inIntSet      : int * IntSet -> bool

      val addIntSet     : int * IntSet -> IntSet
      val delIntSet     : int * IntSet -> IntSet

      val cupIntSets    : IntSet * IntSet -> IntSet
      val capIntSets    : IntSet * IntSet -> IntSet
      val diffIntSets   : IntSet * IntSet -> IntSet

      val IntSet2List : IntSet -> int list
      val IntList2Set : int list -> IntSet

      val toWord :	IntSet -> word
      val toString :	IntSet -> string
   end

structure IntSets : IntSets = 
   struct
      structure W = Word
      val wordSize = W.wordSize
 
      type IntSet = W.word vector
      
      infix 7 << >>
      infix 6 &&
      infix 5 ||

      val op >> = W.>>
      val op << = W.<<
      val op && = W.andb
      val op || = W.orb
      val !! = W.notb

      fun emptyIntSet n = 
	 Vector.tabulate((n+wordSize-1) div wordSize,fn _ => 0w0):IntSet

      fun fullIntSet n = let val size = (n+wordSize-1) div wordSize
			     val full = 0w0-0w1:W.word
			     val bits = (n-1) mod wordSize+1
			     val last = full >> (Word.fromInt (wordSize-bits))
			 in Vector.tabulate(n div wordSize+1,
					    fn i => if i<size-1 then full else last):IntSet
			 end

      fun singleIntSet n = 
	 let 
	    val idx = n div wordSize
	    val mask = 0w1 << (Word.fromInt (n mod wordSize))
	 in 
	    Vector.tabulate(idx+1,fn i => if i=idx then mask else 0w0):IntSet
	 end

      fun isEmptyIntSet vec = Vector.foldr W.orb 0w0 vec = 0w0

      fun inIntSet(n,vec) = 
	 let 
	    val idx = n div wordSize
	    val mask = 0w1 << (Word.fromInt (n mod wordSize))
	 in 
	    Vector.sub(vec,idx) && mask <> 0w0 handle Subscript => false
	 end

      fun eqIntSets(vec1,vec2:IntSet) = 
	 let 
	    val (l1,l2) = (Vector.length vec1,Vector.length vec2)
	    val (l12,v1,v2) = if l1<=l2 then (l1,vec1,vec2) else (l2,vec2,vec1)
	    val eq12 = VectorSlice.foldri 
	       (fn (i,x1,eq) => eq andalso x1=Vector.sub(v2,i)) true (VectorSlice.slice (v1,0,NONE))
	 in VectorSlice.foldri (fn (_,x,eq) => eq andalso x=0w0) eq12 (VectorSlice.slice (v2,l12,NONE))
	 end
      
      fun addIntSet(n,vec) = 
	 let
	    val idx = n div wordSize
	    val mask = 0w1 << (Word.fromInt (n mod wordSize))
	    val size = Vector.length vec
	 in 
	    if size>idx 
	       then VectorSlice.mapi (fn (i,x) => if i=idx then x||mask else x) (VectorSlice.slice (vec,0,NONE))
	    else Vector.tabulate 
	       (idx+1,fn i => if i<size then Vector.sub(vec,i) else if i=idx then mask else 0w0)
	 end

      fun delIntSet(n,vec) = 
	 let
	    val idx = n div wordSize
	    val size = Vector.length vec
	 in 
	    if size<=idx then vec
	    else let val mask = !! (0w1 << (Word.fromInt (n mod wordSize)))
		 in VectorSlice.mapi (fn (i,x) => if i=idx then x && mask else x) (VectorSlice.slice (vec,0,NONE))
		 end
	 end
      
      fun capIntSets(vec1,vec2) = 
	 let 
	    val l12 = Int.min(Vector.length vec1,Vector.length vec2)
	 in 
	    Vector.tabulate(l12,fn i => Vector.sub(vec1,i) && Vector.sub(vec2,i))
	 end

      fun cupIntSets(vec1,vec2) = 
	 let 
	    val (l1,l2) = (Vector.length vec1,Vector.length vec2)
	    val (shorter,longer,v) = if l1<=l2 then (l1,l2,vec2) else (l2,l1,vec1)
	 in 
	    Vector.tabulate (longer,fn i => if i>=shorter then Vector.sub(v,i)
					    else Vector.sub(vec1,i) || Vector.sub(vec2,i))
	 end

      fun diffIntSets(vec1,vec2) = 
	 let 
	    val (l1,l2) = (Vector.length vec1,Vector.length vec2)
	 in 
	    Vector.tabulate (l1,fn i => if i>=l2 then Vector.sub(vec1,i)
					else Vector.sub(vec1,i) && !!(Vector.sub(vec2,i)))
	 end
      
      fun IntList2Set l = let
	val max = foldl Int.max 0 l
	in foldl addIntSet (emptyIntSet max) l
	end

      fun IntSet2List vec = 
	 let 
	    val size = Vector.length vec
	    fun doOne (w,off,yet) = 
	       let fun doit (i,mask) = 
		  if i=wordSize then yet 
		  else if w&&mask=0w0 then doit(i+1,mask<<0wx1)
		       else (off+i)::doit(i+1,mask<<0wx1)
	       in doit(0,0wx1)
	       end
	    fun doAll i = if i>=size then nil
			  else doOne(Vector.sub(vec,i),wordSize*i,(doAll (i+1)))
	 in doAll 0
	 end

      fun super (a,b) = eqIntSets (capIntSets (a,b), b)
      fun toString vec = PrintUtils.intListToString (IntSet2List vec)
      fun toWord vec = Vector.foldl (fn (x,y) => y * 0wx7 + x)  0wx13 vec
   end			      
