signature MutIntSet =
   sig 
      type mutIntSet

      val full    : int -> mutIntSet
      val empty   : int -> mutIntSet
      val single  : int -> mutIntSet
      val copy	  : mutIntSet -> mutIntSet

      val isEmpty : mutIntSet -> bool
      val eq     : mutIntSet * mutIntSet -> bool 
      val super		: mutIntSet * mutIntSet -> bool
      val isIn      : int * mutIntSet -> bool

      val add     : int * mutIntSet -> unit
      val del     : int * mutIntSet -> unit

      val cap    : mutIntSet * mutIntSet -> unit
      val cup    : mutIntSet * mutIntSet -> unit
      val diff   : mutIntSet * mutIntSet -> unit

      val fromIntList : int list -> mutIntSet
      val toIntList : mutIntSet -> int list

      val toIntSet  : mutIntSet -> IntSets.IntSet
      val IntList2Set : int list -> IntSets.IntSet

      val toWord :	mutIntSet -> word
      val toString :	mutIntSet -> string
   end

structure MutIntSet : MutIntSet = 
   struct
      structure W = Word
      val wordSize = W.wordSize
 
      type mutIntSet = W.word array ref
      
      infix 7 << >>
      infix 6 &&
      infix 5 ||

      val op >> = W.>>
      val op << = W.<<
      val op && = W.andb
      val op || = W.orb
      val !! = W.notb

      fun empty n = ref (
	 Array.tabulate((n+wordSize-1) div wordSize,fn _ => 0w0)):mutIntSet

      fun full n = let val size = (n+wordSize-1) div wordSize
			     val full = 0w0-0w1:W.word
			     val bits = (n-1) mod wordSize+1
			     val last = full >> (Word.fromInt (wordSize-bits))
			 in ref (Array.tabulate(n div wordSize+1,
			    fn i => if i<size-1 then full else last)): mutIntSet
			 end

      fun single n = 
	 let 
	    val idx = n div wordSize
	    val mask = 0w1 << (Word.fromInt (n mod wordSize))
	 in ref (
	    	Array.tabulate(idx+1,
		fn i => if i=idx then mask else 0w0)):mutIntSet
	 end

      fun copy (ref vec) = let
	val n = Array.length vec
	in ref (Array.tabulate (n, fn i => Array.sub (vec,i)))
	end

      fun isEmpty (ref vec) = Array.foldr W.orb 0w0 (vec) = 0w0

      fun isIn (n, ref vec) = 
	 let 
	    val idx = n div wordSize
	    val mask = 0w1 << (Word.fromInt (n mod wordSize))
	 in 
	    Array.sub(vec,idx) && mask <> 0w0 handle Subscript => false
	 end

      fun eq (ref vec1, (ref vec2):mutIntSet) = 
	 let 
	    val (l1,l2) = (Array.length vec1,Array.length vec2)
	    val (l12,v1,v2) = if l1<=l2 then (l1,vec1,vec2) else (l2,vec2,vec1)
	    val eq12 = ArraySlice.foldri 
	       (fn (i,x1,eq) => eq andalso x1=Array.sub(v2,i)) true (ArraySlice.slice (v1,0,NONE))
	 in ArraySlice.foldri (fn (_,x,eq) => eq andalso x=0w0) eq12 (ArraySlice.slice (v2,l12,NONE))
	 end
      
      fun add (n,v as ref vec) = 
	 let
	    val idx = n div wordSize
	    val mask = 0w1 << (Word.fromInt (n mod wordSize))
	    val size = Array.length vec
	 in 
	    if size>idx 
	       then ArraySlice.modifyi (fn (i,x) => if i=idx then x||mask else x) (ArraySlice.slice (vec,0,NONE))
	    else v :=
		Array.tabulate (idx + 1,
	        fn i => if i<size then Array.sub(vec,i) 
			else if i=idx then mask else 0w0)
	 end

      fun del (n,ref vec) = 
	 let
	    val idx = n div wordSize
	    val size = Array.length vec
	 in 
	    if size<=idx then ()
	    else let val mask = !! (0w1 << (Word.fromInt (n mod wordSize)))
		 in ArraySlice.modifyi (fn (i,x) => if i=idx then x && mask else x) (ArraySlice.slice (vec,0,NONE))
		 end
	 end
      
      fun cap (ref vec1,ref vec2) = 
	    ArraySlice.modifyi (fn (i,w) => w && Array.sub(vec2,i)) (ArraySlice.slice (vec1,0,NONE))

      fun cup (v as ref vec1,ref vec2) = let
            val (l1,l2) = (Array.length vec1,Array.length vec2)
            in if l1>=l2 then ArraySlice.modifyi (
				fn (i,w) => if i>=l2 then w
                            	else w || Array.sub(vec2,i))
		              (ArraySlice.slice (vec1,0,NONE))
	       else v :=
		Array.tabulate (l2,fn i => if i>=l1 then Array.sub(vec2,i)
                                else Array.sub(vec1,i) || Array.sub(vec2,i))
         end

      fun diff (ref vec1,ref vec2) = 
	    ArraySlice.modifyi (fn (i,w) => w && !!(Array.sub(vec2,i))) (ArraySlice.slice (vec1,0,NONE))
      
      fun super (a,b) = let 
	val b' = copy b
	val _  = cap (b,a)
	in eq (b, b')
	end

      fun fromIntList l = 
	 let 
	    val max = List.foldr Int.max 0 l
	    val result = empty max
	    val _ = app (fn a => add (a,result)) l
	 in 
	    result
	 end

      fun toIntList (ref vec) = 
	 let 
	    val size = Array.length vec
	    fun doOne (w,off,yet) = 
	       let fun doit (i,mask) = 
		  if i=wordSize then yet 
		  else if w&&mask=0w0 then doit(i+1,mask<<0wx1)
		       else (off+i)::doit(i+1,mask<<0wx1)
	       in doit(0,0wx1)
	       end
	    fun doAll i = if i>=size then nil
			  else doOne(Array.sub(vec,i),wordSize*i,(doAll (i+1)))
	 in doAll 0
	 end

      fun toIntSet (ref arr) = Array.vector arr

      val IntList2Set = toIntSet o fromIntList

      fun toString vec = PrintUtils.intListToString (toIntList vec)
      fun toWord (ref vec) = Array.foldl (fn (x,y) => y * 0wx7 + x)  0wx13 vec
   end			      
